"""TypeScript/JavaScript parser: extract functions, classes, methods and build code_id."""

from __future__ import annotations

import fnmatch
import re
from pathlib import Path

from livedoc.core.signatures import CodeEntity

TS_EXTENSIONS = (".ts", ".tsx", ".js", ".jsx")
TS_IGNORE = ("node_modules", "dist", "build", ".git", "*.test.*", "*.spec.*", "*.d.ts")


def _path_to_module(root: Path, file_path: Path) -> str:
    """File path relative to root -> dotted module name (e.g. src.utils.calc)."""
    rel = file_path.relative_to(root)
    parts = list(rel.parts)
    if parts[-1].endswith((".ts", ".tsx", ".js", ".jsx")):
        parts[-1] = parts[-1].rsplit(".", 1)[0]
    return ".".join(parts)


def _normalize_param(param: str) -> str:
    """Extract effective param name from destructuring, rest, etc."""
    param = param.strip()
    if param.startswith("..."):
        return param
    if param.startswith("{"):
        idx = param.find("}")
        inner = (param[1:idx] if idx > 0 else param[1:]).strip()
        if ":" in inner:
            parts = re.split(r"\s*,\s*", inner)
            return parts[0].split(":")[0].strip() if parts else "?"
        return inner.split(",")[0].strip().split(":")[0].strip() if inner else "?"
    if param.startswith("["):
        inner = param[1 : param.rfind("]")].strip()
        return inner.split(",")[0].strip() if inner else "?"
    return param.split(":")[0].strip().split("=")[0].strip()


def _split_top_level_commas(text: str) -> list[str]:
    """Split by commas, ignoring commas inside (), {}, [], <> and strings."""
    parts: list[str] = []
    current: list[str] = []
    paren = brace = bracket = angle = 0
    in_single = False
    in_double = False
    in_backtick = False
    escaped = False

    for ch in text:
        if escaped:
            current.append(ch)
            escaped = False
            continue
        if ch == "\\":
            current.append(ch)
            escaped = True
            continue
        if in_single:
            current.append(ch)
            if ch == "'":
                in_single = False
            continue
        if in_double:
            current.append(ch)
            if ch == '"':
                in_double = False
            continue
        if in_backtick:
            current.append(ch)
            if ch == "`":
                in_backtick = False
            continue
        if ch == "'":
            current.append(ch)
            in_single = True
            continue
        if ch == '"':
            current.append(ch)
            in_double = True
            continue
        if ch == "`":
            current.append(ch)
            in_backtick = True
            continue
        if ch == "(":
            paren += 1
        elif ch == ")":
            paren = max(0, paren - 1)
        elif ch == "{":
            brace += 1
        elif ch == "}":
            brace = max(0, brace - 1)
        elif ch == "[":
            bracket += 1
        elif ch == "]":
            bracket = max(0, bracket - 1)
        elif ch == "<":
            angle += 1
        elif ch == ">":
            angle = max(0, angle - 1)
        elif ch == "," and paren == 0 and brace == 0 and bracket == 0 and angle == 0:
            part = "".join(current).strip()
            if part:
                parts.append(part)
            current = []
            continue
        current.append(ch)

    tail = "".join(current).strip()
    if tail:
        parts.append(tail)
    return parts


def _find_matching_paren(source: str, open_idx: int) -> int:
    """Return index of matching ')' for source[open_idx] == '('; -1 if not found."""
    depth = 0
    in_single = in_double = in_backtick = False
    escaped = False
    for i in range(open_idx, len(source)):
        ch = source[i]
        if escaped:
            escaped = False
            continue
        if ch == "\\":
            escaped = True
            continue
        if in_single:
            if ch == "'":
                in_single = False
            continue
        if in_double:
            if ch == '"':
                in_double = False
            continue
        if in_backtick:
            if ch == "`":
                in_backtick = False
            continue
        if ch == "'":
            in_single = True
            continue
        if ch == '"':
            in_double = True
            continue
        if ch == "`":
            in_backtick = True
            continue
        if ch == "(":
            depth += 1
        elif ch == ")":
            depth -= 1
            if depth == 0:
                return i
    return -1


def _return_before(source: str, start: int, stop_char: str) -> str:
    """Extract ': Type' between start and next stop_char, if present."""
    end = source.find(stop_char, start)
    if end == -1:
        return ""
    between = source[start:end].strip()
    if between.startswith(":"):
        return between[1:].strip()
    return ""


def _extract_args_from_parens(s: str) -> list[str]:
    """Extract param names from function(params) string, handling nested parens."""
    args: list[str] = []
    depth = 0
    start = -1
    for i, c in enumerate(s):
        if c == "(":
            if depth == 0:
                start = i
            depth += 1
        elif c == ")":
            depth -= 1
            if depth == 0 and start >= 0:
                inner = s[start + 1 : i]
                for part in re.split(r"\s*,\s*", inner):
                    part = part.strip()
                    if not part or part.startswith("..."):
                        continue
                    # name: type or just name
                    name = part.split(":")[0].strip().split("=")[0].strip()
                    if name and name not in ("this", "context"):
                        args.append(name)
                break
    return args


def _find_functions(source: str, module_path: str, file_path: Path) -> list[CodeEntity]:
    """Find function declarations: function name(...) and const name = (...) =>."""
    entities: list[CodeEntity] = []
    # function name(...) or export/default/async variants
    pattern = r"(?:export\s+(?:default\s+)?)?(?:async\s+)?function\s+(\w+)(?:\s*<[^>]*>)?\s*\("
    for m in re.finditer(pattern, source):
        name = m.group(1)
        open_idx = m.end() - 1
        close_idx = _find_matching_paren(source, open_idx)
        if close_idx == -1:
            continue
        args_str = source[open_idx + 1 : close_idx]
        ret = _return_before(source, close_idx + 1, "{")
        args = [_normalize_param(p) for p in _split_top_level_commas(args_str) if p.strip()]
        line = source[: m.start()].count("\n") + 1
        entities.append(
            CodeEntity(
                code_id=f"{module_path}:{name}",
                name=name,
                args=args,
                return_annotation=ret,
                file_path=file_path,
                line=line,
            )
        )
    # const/let name = (params) => or const/let name = (params): Type =>
    pattern2 = r"(?:export\s+)?(?:const|let)\s+(\w+)\s*=\s*(?:async\s+)?\("
    for m in re.finditer(pattern2, source):
        name = m.group(1)
        open_idx = m.end() - 1
        close_idx = _find_matching_paren(source, open_idx)
        if close_idx == -1:
            continue
        args_str = source[open_idx + 1 : close_idx]
        after = source[close_idx + 1 :]
        arrow_pos = after.find("=>")
        ret = ""
        if arrow_pos != -1:
            between = after[:arrow_pos].strip()
            if between.startswith(":"):
                ret = between[1:].strip()
        args = [_normalize_param(p) for p in _split_top_level_commas(args_str) if p.strip()]
        line = source[: m.start()].count("\n") + 1
        entities.append(
            CodeEntity(
                code_id=f"{module_path}:{name}",
                name=name,
                args=args,
                return_annotation=ret,
                file_path=file_path,
                line=line,
            )
        )
    return entities


def _find_classes_and_methods(source: str, module_path: str, file_path: Path) -> list[CodeEntity]:
    """Find class declarations and their methods."""
    entities: list[CodeEntity] = []
    # class Name { ... } or export default class Name { ... }
    class_pattern = r"(?:export\s+(?:default\s+)?)?class\s+(\w+)(?:\s+extends\s+\w+)?\s*\{"
    for class_match in re.finditer(class_pattern, source):
        class_name = class_match.group(1)
        start = class_match.end()
        brace_depth = 1
        i = start
        while i < len(source) and brace_depth > 0:
            if source[i] == "{":
                brace_depth += 1
            elif source[i] == "}":
                brace_depth -= 1
            i += 1
        class_body = source[start:i]
        # Find methods: name(...) { or async name(...) { (inside class body)
        method_pattern = r"(?:async\s+)?(\w+)\s*\("
        for m in re.finditer(method_pattern, class_body):
            name = m.group(1)
            if name == "constructor":
                continue
            open_idx = class_match.end() + m.end() - 1
            close_idx = _find_matching_paren(source, open_idx)
            if close_idx == -1:
                continue
            args_str = source[open_idx + 1 : close_idx]
            ret = _return_before(source, close_idx + 1, "{")
            args = [_normalize_param(p) for p in _split_top_level_commas(args_str) if p.strip()]
            if "this" in args:
                args = [a for a in args if a != "this"]
            line = source[: m.start() + class_match.end()].count("\n") + 1
            entities.append(
                CodeEntity(
                    code_id=f"{module_path}:{class_name}.{name}",
                    name=name,
                    args=args,
                    return_annotation=ret,
                    file_path=file_path,
                    line=line,
                )
            )
    return entities


def _find_interfaces(source: str, module_path: str, file_path: Path) -> list[CodeEntity]:
    """Find interface declarations: interface Name { ... }."""
    entities: list[CodeEntity] = []
    pattern = r"(?:export\s+)?interface\s+(\w+)(?:\s+extends\s+[\w\s,]+)?\s*\{"
    for m in re.finditer(pattern, source):
        name = m.group(1)
        start = m.end()
        brace_depth = 1
        i = start
        while i < len(source) and brace_depth > 0:
            if source[i] == "{":
                brace_depth += 1
            elif source[i] == "}":
                brace_depth -= 1
            i += 1
        body = source[start : i - 1]
        # Extract property names: "name: Type" or "readonly name: Type" or "name?: Type"
        props = re.findall(r"(?:readonly\s+)?(\w+)\s*\??\s*[:=]", body)
        line = source[: m.start()].count("\n") + 1
        entities.append(
            CodeEntity(
                code_id=f"{module_path}:{name}",
                name=name,
                args=props,
                return_annotation="interface",
                file_path=file_path,
                line=line,
            )
        )
    return entities


def _find_type_aliases(source: str, module_path: str, file_path: Path) -> list[CodeEntity]:
    """Find type alias declarations: type Name = ..."""
    entities: list[CodeEntity] = []
    # type Name = ... (until ; or newline or next declaration)
    pattern = r"(?:export\s+)?type\s+(\w+)\s*=\s*"
    for m in re.finditer(pattern, source):
        name = m.group(1)
        rest = source[m.end() :]
        # Extract type expression: balance braces/parens, stop at ; or \n\n or next keyword
        depth = 0
        in_brace, in_paren, in_bracket = 0, 0, 0
        i = 0
        while i < len(rest):
            c = rest[i]
            if c == "{":
                in_brace += 1
            elif c == "}":
                in_brace -= 1
            elif c == "(":
                in_paren += 1
            elif c == ")":
                in_paren -= 1
            elif c == "[":
                in_bracket += 1
            elif c == "]":
                in_bracket -= 1
            elif c == ";" and in_brace == in_paren == in_bracket == 0:
                break
            elif c == "\n" and in_brace == in_paren == in_bracket == 0:
                break
            i += 1
        type_expr = rest[:i].strip()
        type_expr = re.sub(r"\s+", " ", type_expr)[:200]
        line = source[: m.start()].count("\n") + 1
        entities.append(
            CodeEntity(
                code_id=f"{module_path}:{name}",
                name=name,
                args=[],
                return_annotation=type_expr,
                file_path=file_path,
                line=line,
            )
        )
    return entities


def parse_typescript_file(file_path: Path, module_path: str) -> list[CodeEntity]:
    """Parse one .ts/.tsx/.js/.jsx file and return entities."""
    source = file_path.read_text(encoding="utf-8")
    entities: list[CodeEntity] = []
    entities.extend(_find_functions(source, module_path, file_path))
    entities.extend(_find_classes_and_methods(source, module_path, file_path))
    entities.extend(_find_interfaces(source, module_path, file_path))
    entities.extend(_find_type_aliases(source, module_path, file_path))
    return entities


def _is_ignored(rel_path: Path, ignore_patterns: tuple[str, ...]) -> bool:
    """Return True if path matches any ignore pattern."""
    parts = rel_path.parts
    for pattern in ignore_patterns:
        for part in parts:
            if fnmatch.fnmatch(part, pattern) or part == pattern:
                return True
    if any(fnmatch.fnmatch(rel_path.name, p) for p in ("*.test.*", "*.spec.*", "*.d.ts")):
        return True
    return False


def parse_typescript_module(
    root: Path,
    package_path: Path,
    ignore_patterns: tuple[str, ...] = (),
) -> list[CodeEntity]:
    """
    Recursively parse TypeScript/JavaScript files and return all entities.
    """
    ignore = list(TS_IGNORE) + list(ignore_patterns)
    all_entities: list[CodeEntity] = []
    for ext in TS_EXTENSIONS:
        for path in package_path.rglob(f"*{ext}"):
            try:
                rel = path.relative_to(root)
            except ValueError:
                rel = path
            if _is_ignored(rel, tuple(ignore)):
                continue
            module_path = _path_to_module(root, path)
            all_entities.extend(parse_typescript_file(path, module_path))
    return all_entities



"""Go parser: extract functions, methods and build code_id."""

from __future__ import annotations

import fnmatch
import re
from pathlib import Path

from livedoc.core.signatures import CodeEntity

GO_EXTENSIONS = (".go",)
GO_IGNORE = ("vendor", ".git", "*_test.go")


def _extract_package(source: str) -> str:
    """Extract package name from 'package name' declaration."""
    m = re.search(r"^package\s+(\w+)", source, re.MULTILINE)
    return m.group(1) if m else "main"


def _split_go_top_level_commas(text: str) -> list[str]:
    """Split Go params by top-level commas (ignore commas in (), [], {})."""
    parts: list[str] = []
    current: list[str] = []
    paren = brace = bracket = 0
    for ch in text:
        if ch == "(":
            paren += 1
        elif ch == ")":
            paren = max(0, paren - 1)
        elif ch == "{":
            brace += 1
        elif ch == "}":
            brace = max(0, brace - 1)
        elif ch == "[":
            bracket += 1
        elif ch == "]":
            bracket = max(0, bracket - 1)
        elif ch == "," and paren == 0 and brace == 0 and bracket == 0:
            part = "".join(current).strip()
            if part:
                parts.append(part)
            current = []
            continue
        current.append(ch)
    tail = "".join(current).strip()
    if tail:
        parts.append(tail)
    return parts


def _find_matching_paren(source: str, open_idx: int) -> int:
    """Return index of matching ')' for source[open_idx] == '('; -1 if not found."""
    depth = 0
    for i in range(open_idx, len(source)):
        ch = source[i]
        if ch == "(":
            depth += 1
        elif ch == ")":
            depth -= 1
            if depth == 0:
                return i
    return -1


def _return_before(source: str, start: int, stop_char: str) -> str:
    end = source.find(stop_char, start)
    if end == -1:
        return ""
    return source[start:end].strip()


def _extract_go_params(params_str: str) -> list[str]:
    """Extract param names from Go params: 'a, b int' or 'ctx context.Context'."""
    args: list[str] = []
    # Split by comma at top level (Go allows "a, b int" - names share type)
    parts = _split_go_top_level_commas(params_str.strip())
    pending_names: list[str] = []
    for part in parts:
        part = part.strip()
        if not part:
            continue
        if part.startswith("..."):
            # Unnamed variadic parameter: type-only, skip.
            continue
        tokens = part.split()
        if len(tokens) >= 2:
            name = tokens[0]
            if re.fullmatch(r"[A-Za-z_]\w*", name) and name != "_":
                args.extend(pending_names)
                pending_names = []
                args.append(name)
            else:
                pending_names = []
            continue
        token = tokens[0]
        if re.fullmatch(r"[A-Za-z_]\w*", token) and token != "_":
            # Could be "a" from "a, b int" or unnamed basic type.
            # Keep temporarily; flush only when a typed segment appears.
            pending_names.append(token)
        else:
            pending_names = []
    return args


def parse_go_file(file_path: Path, pkg: str) -> list[CodeEntity]:
    """Parse one .go file and return entities."""
    source = file_path.read_text(encoding="utf-8")
    entities: list[CodeEntity] = []
    # Methods first (they match func (r) Name pattern)
    for m in re.finditer(r"\bfunc\s+\(([^)]+)\)\s+(\w+)\s*\(", source):
        recv = m.group(1).strip().split()
        method_name = m.group(2)
        open_idx = m.end() - 1
        close_idx = _find_matching_paren(source, open_idx)
        if close_idx == -1:
            continue
        params_str = source[open_idx + 1 : close_idx].strip()
        ret = _return_before(source, close_idx + 1, "{")
        if not recv:
            continue
        type_name = recv[-1]
        qualified = f"(*{type_name[1:]}).{method_name}" if type_name.startswith("*") else f"{type_name}.{method_name}"
        args = _extract_go_params(params_str)
        line = source[: m.start()].count("\n") + 1
        entities.append(
            CodeEntity(
                code_id=f"{pkg}:{qualified}",
                name=method_name,
                args=args,
                return_annotation=ret,
                file_path=file_path,
                line=line,
            )
        )
    # Functions (methods are excluded by requiring "func Name(" pattern)
    func_pattern = r"\bfunc\s+(\w+)\s*\("
    for m in re.finditer(func_pattern, source):
        name = m.group(1)
        open_idx = m.end() - 1
        close_idx = _find_matching_paren(source, open_idx)
        if close_idx == -1:
            continue
        params_str = source[open_idx + 1 : close_idx].strip()
        ret = _return_before(source, close_idx + 1, "{")
        args = _extract_go_params(params_str)
        line = source[: m.start()].count("\n") + 1
        entities.append(
            CodeEntity(
                code_id=f"{pkg}:{name}",
                name=name,
                args=args,
                return_annotation=ret,
                file_path=file_path,
                line=line,
            )
        )
    return entities


def _is_ignored(rel_path: Path, ignore_patterns: tuple[str, ...]) -> bool:
    """Return True if path matches any ignore pattern."""
    parts = rel_path.parts
    for pattern in ignore_patterns:
        for part in parts:
            if fnmatch.fnmatch(part, pattern) or part == pattern:
                return True
    if fnmatch.fnmatch(rel_path.name, "*_test.go"):
        return True
    return False


def parse_go_module(
    root: Path,
    package_path: Path,
    ignore_patterns: tuple[str, ...] = (),
) -> list[CodeEntity]:
    """Recursively parse Go files and return all entities."""
    ignore = list(GO_IGNORE) + list(ignore_patterns)
    all_entities: list[CodeEntity] = []
    for path in package_path.rglob("*.go"):
        try:
            rel = path.relative_to(root)
        except ValueError:
            rel = path
        if _is_ignored(rel, tuple(ignore)):
            continue
        source = path.read_text(encoding="utf-8")
        pkg = _extract_package(source)
        all_entities.extend(parse_go_file(path, pkg))
    return all_entities



"""Load config from .livedoc.json in project root."""

from __future__ import annotations

import json
from pathlib import Path


CONFIG_FILE = ".livedoc.json"


def load_config(root: Path) -> dict:
    """
    Load config from root/.livedoc.json.
    Returns dict with keys: docs, ignore, format. Missing keys use defaults.
    """
    path = root.resolve() / CONFIG_FILE
    if not path.exists():
        return {}
    try:
        data = json.loads(path.read_text(encoding="utf-8"))
        if not isinstance(data, dict):
            return {}
        result: dict = {}
        if "docs" in data and isinstance(data["docs"], str):
            result["docs"] = data["docs"]
        if "ignore" in data and isinstance(data["ignore"], list):
            result["ignore"] = [str(x) for x in data["ignore"] if isinstance(x, str)]
        if "format" in data and data["format"] in ("text", "json"):
            result["format"] = data["format"]
        return result
    except (json.JSONDecodeError, OSError):
        return {}



"""Tests for parsers and outdated detection."""

from pathlib import Path

import pytest

from livedoc.core.graph import DocGraph
from livedoc.core.signatures import CodeSignatures, signature_hash
from livedoc.parsers.doc_parser import parse_doc_file
from livedoc.parsers.python_parser import (
    build_current_signatures,
    parse_python_file,
    parse_python_module,
)
from livedoc.parsers.typescript_parser import (
    parse_typescript_file,
    parse_typescript_module,
)
from livedoc.parsers.go_parser import parse_go_file, parse_go_module


EXAMPLES_ROOT = Path(__file__).resolve().parent.parent / "examples"


def test_signature_hash_stable() -> None:
    h1 = signature_hash("add", ["a", "b"], "int")
    h2 = signature_hash("add", ["a", "b"], "int")
    assert h1 == h2


def test_signature_hash_changes_with_args() -> None:
    h1 = signature_hash("add", ["a", "b"], "int")
    h2 = signature_hash("add", ["a", "b", "c"], "int")
    assert h1 != h2


def test_parse_python_file() -> None:
    calc = EXAMPLES_ROOT / "sample_module" / "calc.py"
    root = EXAMPLES_ROOT
    module_path = "sample_module.calc"
    entities = parse_python_file(calc, module_path)
    code_ids = {e.code_id for e in entities}
    assert "sample_module.calc:add" in code_ids
    assert "sample_module.calc:subtract" in code_ids
    assert "sample_module.calc:Calculator.multiply" in code_ids
    assert "sample_module.calc:Calculator.divide" in code_ids


def test_parse_doc_file() -> None:
    api_md = EXAMPLES_ROOT / "docs" / "api.md"
    docs_root = EXAMPLES_ROOT / "docs"
    fragments = parse_doc_file(api_md, docs_root)
    assert len(fragments) >= 3
    code_ids_found = set()
    for f in fragments:
        code_ids_found.update(f.code_ids)
    assert "sample_module.calc:add" in code_ids_found
    assert "sample_module.calc:Calculator.multiply" in code_ids_found


def test_doc_graph_outdated() -> None:
    from livedoc.core.graph import DocFragment

    graph = DocGraph()
    f1 = DocFragment("api.md#add", Path("api.md"), 1, ["m:add"], "add")
    f2 = DocFragment("api.md#sub", Path("api.md"), 2, ["m:subtract"], "subtract")
    graph.add_link("m:add", f1)
    graph.add_link("m:subtract", f2)
    outdated = graph.get_outdated_fragments({"m:add"})
    assert len(outdated) == 1
    assert outdated[0].doc_fragment_id == "api.md#add"


def test_code_signatures_changed() -> None:
    stored = CodeSignatures({"m:add": "hash1", "m:sub": "hash2"})
    current = {"m:add": "hash1", "m:sub": "hash3"}
    changed = stored.changed_code_ids(current)
    assert changed == {"m:sub"}


def test_code_signatures_removed() -> None:
    stored = CodeSignatures({"m:add": "hash1"})
    current: dict[str, str] = {}
    changed = stored.changed_code_ids(current)
    assert changed == {"m:add"}


def test_parse_python_module_ignores_tests() -> None:
    """Default ignore excludes tests/ and similar paths."""
    from livedoc.parsers.python_parser import DEFAULT_IGNORE, parse_python_module

    root = EXAMPLES_ROOT
    entities = parse_python_module(root, root, ignore_patterns=DEFAULT_IGNORE)
    # examples/ has sample_module, no tests - so we get calc entities
    code_ids = {e.code_id for e in entities}
    assert "sample_module.calc:add" in code_ids
    # With custom ignore including sample_module, we'd get nothing
    empty = parse_python_module(root, root, ignore_patterns=("sample_module",))
    assert len(empty) == 0


def test_load_livedocignore() -> None:
    from livedoc.cli import _load_livedocignore

    import tempfile
    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        ignore_file = root / ".livedocignore"
        ignore_file.write_text("build\n# comment\nscripts\n", encoding="utf-8")
        patterns = _load_livedocignore(root)
        assert "build" in patterns
        assert "scripts" in patterns
        assert "#" not in "".join(patterns)


def test_run_check_quiet_no_output(tmp_path: Path, capsys: pytest.CaptureFixture[str]) -> None:
    from livedoc.cli import run_check

    (tmp_path / "mod.py").write_text(
        "def add(a: int, b: int) -> int:\n    return a + b\n",
        encoding="utf-8",
    )
    docs_dir = tmp_path / "docs"
    docs_dir.mkdir()
    (docs_dir / "api.md").write_text(
        '<!-- livedoc: code_id = "mod:add" -->\n## add\nAdd docs.\n',
        encoding="utf-8",
    )

    rc_first = run_check(tmp_path, "docs", update_signatures=False, quiet=True)
    out_first = capsys.readouterr()
    assert rc_first == 0
    assert out_first.out == ""

    rc_second = run_check(tmp_path, "docs", update_signatures=False, quiet=True)
    out_second = capsys.readouterr()
    assert rc_second == 0
    assert out_second.out == ""


def test_run_check_quiet_still_prints_failures(tmp_path: Path, capsys: pytest.CaptureFixture[str]) -> None:
    from livedoc.cli import run_check

    mod_file = tmp_path / "mod.py"
    mod_file.write_text(
        "def add(a: int, b: int) -> int:\n    return a + b\n",
        encoding="utf-8",
    )
    docs_dir = tmp_path / "docs"
    docs_dir.mkdir()
    (docs_dir / "api.md").write_text(
        '<!-- livedoc: code_id = "mod:add" -->\n## add\nAdd docs.\n',
        encoding="utf-8",
    )

    assert run_check(tmp_path, "docs", update_signatures=False, quiet=True) == 0
    _ = capsys.readouterr()

    mod_file.write_text(
        "def add(a: int, b: int, c: int) -> int:\n    return a + b + c\n",
        encoding="utf-8",
    )
    rc = run_check(tmp_path, "docs", update_signatures=False, quiet=True)
    out = capsys.readouterr()
    assert rc == 1
    assert "Possibly outdated documentation" in out.out
    assert "mod:add" in out.out


def test_report_outdated_with_changes() -> None:
    import tempfile
    from livedoc.core.graph import DocFragment
    from livedoc.core.signatures import CodeEntity
    from livedoc.report.reporter import report_outdated

    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        code_file = root / "calc.py"
        code_file.write_text("def add(): pass\n", encoding="utf-8")
        f = DocFragment("api.md#add", Path("api.md"), 5, ["m:add"], "add")
        changes = {"m:add": ("add(a, b) -> int", "add(a, b, c) -> int")}
        entities = {
            "m:add": CodeEntity(
                code_id="m:add",
                name="add",
                args=["a", "b", "c"],
                return_annotation="int",
                file_path=code_file,
                line=12,
            )
        }
        report = report_outdated([f], changes=changes, entities_by_id=entities, project_root=root)
    assert "add(a, b) -> int" in report
    assert "add(a, b, c) -> int" in report
    assert "m:add" in report
    assert "calc.py:12" in report
    assert "Code:" in report


def test_report_outdated_json_format() -> None:
    import tempfile
    from livedoc.core.graph import DocFragment
    from livedoc.core.signatures import CodeEntity
    from livedoc.report.reporter import report_outdated

    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        cf = root / "m.py"
        cf.write_text("x", encoding="utf-8")
        f = DocFragment("api.md#add", Path("api.md"), 5, ["m:add"], "add")
        changes = {"m:add": ("add(a)", "add(a, b)")}
        entities = {
            "m:add": CodeEntity(
                code_id="m:add",
                name="add",
                args=["a", "b"],
                return_annotation="",
                file_path=cf,
                line=7,
            )
        }
        report = report_outdated(
            [f],
            changes=changes,
            entities_by_id=entities,
            project_root=root,
            output_format="json",
        )
    assert '"ok": false' in report
    assert "api.md#add" in report
    assert "m:add" in report
    assert '"code_file": "m.py"' in report
    assert '"code_line": 7' in report


def test_report_up_to_date_json() -> None:
    from livedoc.report.reporter import report_outdated

    report = report_outdated([], output_format="json")
    assert '"ok": true' in report
    assert '"outdated": []' in report
    assert '"unknown_anchors": []' in report


def test_find_unknown_anchor_refs() -> None:
    from livedoc.core.graph import DocFragment, find_unknown_anchor_refs

    f = DocFragment("api.md#x", Path("api.md"), 1, ["known:id", "missing:func"], "x")
    refs = find_unknown_anchor_refs([f], {"known:id"})
    assert len(refs) == 1
    assert refs[0][0] == "missing:func"
    assert refs[0][1] == f


def test_report_unknown_anchors_text() -> None:
    from livedoc.core.graph import DocFragment
    from livedoc.report.reporter import report_outdated

    f = DocFragment("api.md#bad", Path("api.md"), 3, ["nope:missing"], "Bad")
    report = report_outdated([], unknown_refs=[("nope:missing", f)])
    assert "Unknown code_id" in report
    assert "nope:missing" in report
    assert "api.md" in report


def test_report_outdated_removed_symbol_shows_hint() -> None:
    from livedoc.core.graph import DocFragment
    from livedoc.report.reporter import report_outdated

    f = DocFragment("api.md#gone", Path("api.md"), 1, ["m:gone"], "gone")
    changes = {"m:gone": ("gone() -> void", None)}
    report = report_outdated([f], changes=changes, entities_by_id={})
    assert "removed from codebase" in report


def test_report_unknown_anchors_json() -> None:
    from livedoc.core.graph import DocFragment
    from livedoc.report.reporter import report_outdated

    f = DocFragment("api.md#bad", Path("api.md"), 3, ["x"], "Bad")
    report = report_outdated([], unknown_refs=[("ghost:fn", f)], output_format="json")
    assert '"ok": false' in report
    assert "ghost:fn" in report
    assert "unknown_anchors" in report


def test_config_load() -> None:
    import tempfile
    from livedoc.config import load_config

    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        config_path = root / ".livedoc.json"
        config_path.write_text(
            '{"docs": "mydocs", "ignore": ["build"], "format": "json"}',
            encoding="utf-8",
        )
        config = load_config(root)
        assert config["docs"] == "mydocs"
        assert config["ignore"] == ["build"]
        assert config["format"] == "json"


def test_config_load_empty_missing() -> None:
    from livedoc.config import load_config

    config = load_config(Path("/nonexistent"))
    assert config == {}


def test_config_load_invalid_json() -> None:
    import tempfile
    from livedoc.config import load_config

    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        (root / ".livedoc.json").write_text("not json {", encoding="utf-8")
        config = load_config(root)
        assert config == {}


def test_code_entity_format_signature() -> None:
    from livedoc.core.signatures import CodeEntity

    e = CodeEntity(
        code_id="m:add",
        name="add",
        args=["a", "b"],
        return_annotation="int",
        file_path=Path("x.py"),
        line=1,
    )
    assert e.format_signature() == "add(a, b) -> int"


def test_parse_doc_file_empty() -> None:
    import tempfile
    from livedoc.parsers.doc_parser import parse_doc_file

    with tempfile.NamedTemporaryFile(mode="w", suffix=".md", delete=False) as f:
        f.write("# No anchors here\n")
        path = Path(f.name)
    try:
        fragments = parse_doc_file(path, path.parent)
        assert fragments == []
    finally:
        path.unlink()


def test_parse_typescript_file() -> None:
    utils = EXAMPLES_ROOT / "ts_sample" / "utils.ts"
    root = EXAMPLES_ROOT
    module_path = "ts_sample.utils"
    entities = parse_typescript_file(utils, module_path)
    code_ids = {e.code_id for e in entities}
    assert "ts_sample.utils:add" in code_ids
    assert "ts_sample.utils:subtract" in code_ids
    assert "ts_sample.utils:multiply" in code_ids
    assert "ts_sample.utils:Calculator.divide" in code_ids
    assert "ts_sample.utils:createCalculator" in code_ids
    assert "ts_sample.utils:Point" in code_ids
    assert "ts_sample.utils:UserId" in code_ids
    assert "ts_sample.utils:greet" in code_ids
    point = next(e for e in entities if e.code_id == "ts_sample.utils:Point")
    assert "x" in point.args and "y" in point.args
    greet_ent = next(e for e in entities if e.code_id == "ts_sample.utils:greet")
    assert "name" in greet_ent.args


def test_parse_go_file() -> None:
    calc = EXAMPLES_ROOT / "go_sample" / "calc.go"
    entities = parse_go_file(calc, "calc")
    code_ids = {e.code_id for e in entities}
    assert "calc:Add" in code_ids
    assert "calc:Subtract" in code_ids
    assert "calc:(*Calculator).Multiply" in code_ids
    assert "calc:Calculator.Divide" in code_ids
    add_ent = next(e for e in entities if e.code_id == "calc:Add")
    assert "a" in add_ent.args and "b" in add_ent.args


def test_parse_go_module() -> None:
    root = EXAMPLES_ROOT
    entities = parse_go_module(root, root, ignore_patterns=())
    code_ids = {e.code_id for e in entities}
    assert "calc:Add" in code_ids
    assert "calc:(*Calculator).Multiply" in code_ids


def test_parse_go_ignores_test_files() -> None:
    """*_test.go files should be ignored."""
    import tempfile
    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        (root / "calc.go").write_text(
            "package calc\n\nfunc Add(a, b int) int { return a + b }\n",
            encoding="utf-8",
        )
        (root / "calc_test.go").write_text(
            "package calc\n\nfunc TestAdd(t *testing.T) {}\n",
            encoding="utf-8",
        )
        entities = parse_go_module(root, root, ignore_patterns=())
        code_ids = {e.code_id for e in entities}
        assert "calc:Add" in code_ids
        assert "calc:TestAdd" not in code_ids


def test_parse_go_edge_no_return_and_variadic() -> None:
    import tempfile

    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        src = root / "edge.go"
        src.write_text(
            (
                "package edge\n\n"
                "func Log(msg string, tags ...string) {\n"
                "}\n\n"
                "type Logger struct{}\n\n"
                "func (l *Logger) Close() {\n"
                "}\n"
            ),
            encoding="utf-8",
        )
        entities = parse_go_module(root, root, ignore_patterns=())
        by_id = {e.code_id: e for e in entities}
        assert "edge:Log" in by_id
        assert by_id["edge:Log"].args == ["msg", "tags"]
        assert by_id["edge:Log"].return_annotation == ""
        assert "edge:(*Logger).Close" in by_id
        assert by_id["edge:(*Logger).Close"].return_annotation == ""


def test_parse_go_edge_callback_and_multireturn() -> None:
    import tempfile

    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        src = root / "advanced.go"
        src.write_text(
            (
                "package advanced\n\n"
                "func Transform(items []int, mapper func(a int, b int) int) ([]int, error) {\n"
                "    return nil, nil\n"
                "}\n\n"
                "type Service struct{}\n\n"
                "func (s Service) Merge(in map[string][]int) (map[string]int, error) {\n"
                "    return nil, nil\n"
                "}\n"
            ),
            encoding="utf-8",
        )
        entities = parse_go_module(root, root, ignore_patterns=())
        by_id = {e.code_id: e for e in entities}
        assert "advanced:Transform" in by_id
        assert by_id["advanced:Transform"].args == ["items", "mapper"]
        assert by_id["advanced:Transform"].return_annotation == "([]int, error)"
        assert "advanced:Service.Merge" in by_id
        assert by_id["advanced:Service.Merge"].args == ["in"]
        assert by_id["advanced:Service.Merge"].return_annotation == "(map[string]int, error)"


def test_parse_go_edge_grouped_names_channels_and_unnamed_types() -> None:
    import tempfile

    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        src = root / "channels.go"
        src.write_text(
            (
                "package channels\n\n"
                "func Sum(a, b, c int) int { return a + b + c }\n\n"
                "func Pipe(in <-chan int, out chan<- int, mapper func(a int, b int) int) {\n"
                "}\n\n"
                "func Consume(int, string) {\n"
                "}\n"
            ),
            encoding="utf-8",
        )
        entities = parse_go_module(root, root, ignore_patterns=())
        by_id = {e.code_id: e for e in entities}
        assert "channels:Sum" in by_id
        assert by_id["channels:Sum"].args == ["a", "b", "c"]
        assert "channels:Pipe" in by_id
        assert by_id["channels:Pipe"].args == ["in", "out", "mapper"]
        assert "channels:Consume" in by_id
        assert by_id["channels:Consume"].args == []


def test_parse_go_edge_blank_identifier_and_context_arg() -> None:
    import tempfile

    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        src = root / "ctx.go"
        src.write_text(
            (
                "package ctxpkg\n\n"
                "import \"context\"\n\n"
                "func Handle(ctx context.Context, _ string, value int) error {\n"
                "    return nil\n"
                "}\n"
            ),
            encoding="utf-8",
        )
        entities = parse_go_module(root, root, ignore_patterns=())
        by_id = {e.code_id: e for e in entities}
        assert "ctxpkg:Handle" in by_id
        assert by_id["ctxpkg:Handle"].args == ["ctx", "value"]
        assert by_id["ctxpkg:Handle"].return_annotation == "error"


def test_parse_typescript_ignores_d_ts() -> None:
    """*.d.ts files should be ignored."""
    import tempfile
    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        (root / "types.d.ts").write_text(
            "export declare function foo(x: number): void;\n",
            encoding="utf-8",
        )
        (root / "impl.ts").write_text(
            "export function bar(): number { return 1; }\n",
            encoding="utf-8",
        )
        entities = parse_typescript_module(root, root, ignore_patterns=())
        code_ids = {e.code_id for e in entities}
        assert "impl:bar" in code_ids
        assert "types:foo" not in code_ids


def test_parse_typescript_module() -> None:
    root = EXAMPLES_ROOT
    entities = parse_typescript_module(root, root, ignore_patterns=())
    code_ids = {e.code_id for e in entities}
    assert "ts_sample.utils:add" in code_ids
    assert "ts_sample.utils:Calculator.divide" in code_ids


def test_parse_typescript_edge_params() -> None:
    import tempfile

    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        src = root / "edge.ts"
        src.write_text(
            (
                "export function withDefaults(a: number = 1, opts = { x: 1, y: 2 }): void {}\n"
                "export const withArrow = ({ name, age }: { name: string; age: number }, tags: string[] = [\"a\", \"b\"]) => tags.length\n"
            ),
            encoding="utf-8",
        )
        entities = parse_typescript_module(root, root, ignore_patterns=())
        by_id = {e.code_id: e for e in entities}
        assert "edge:withDefaults" in by_id
        assert "edge:withArrow" in by_id
        assert by_id["edge:withDefaults"].args == ["a", "opts"]
        assert by_id["edge:withArrow"].args == ["name", "tags"]


def test_parse_typescript_edge_generics_and_callback_types() -> None:
    import tempfile

    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        src = root / "generic.ts"
        src.write_text(
            (
                "export function identity<T>(value: T): T { return value; }\n"
                "export function withCallback(cb: (x: number, y: number) => void): void { cb(1, 2); }\n"
                "export const reduceValues = (mapper: (x: number, y: number) => number, initial = 0): number => initial\n"
            ),
            encoding="utf-8",
        )
        entities = parse_typescript_module(root, root, ignore_patterns=())
        by_id = {e.code_id: e for e in entities}
        assert "generic:identity" in by_id
        assert "generic:withCallback" in by_id
        assert "generic:reduceValues" in by_id
        assert by_id["generic:identity"].args == ["value"]
        assert by_id["generic:withCallback"].args == ["cb"]
        assert by_id["generic:reduceValues"].args == ["mapper", "initial"]
        assert by_id["generic:reduceValues"].return_annotation == "number"


def test_parse_typescript_edge_multiline_and_class_callback_method() -> None:
    import tempfile

    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        src = root / "multi.ts"
        src.write_text(
            (
                "export default async function fetchMapped<T>(\n"
                "  opts: { url: string; headers?: Record<string, string> } = { url: \"/\" },\n"
                "  mapper: (x: T, y: number) => Promise<T>,\n"
                "): Promise<T> {\n"
                "  return mapper({} as T, 1)\n"
                "}\n\n"
                "export class Service {\n"
                "  run(\n"
                "    handler: (a: number, b: number) => number,\n"
                "    values: Array<number>,\n"
                "  ): number {\n"
                "    return handler(values[0], values[1])\n"
                "  }\n"
                "}\n"
            ),
            encoding="utf-8",
        )
        entities = parse_typescript_module(root, root, ignore_patterns=())
        by_id = {e.code_id: e for e in entities}
        assert "multi:fetchMapped" in by_id
        assert by_id["multi:fetchMapped"].args == ["opts", "mapper"]
        assert "Promise<T>" in by_id["multi:fetchMapped"].return_annotation
        assert "multi:Service.run" in by_id
        assert by_id["multi:Service.run"].args == ["handler", "values"]


def test_parse_typescript_edge_interface_extends_and_overload_impl() -> None:
    import tempfile

    with tempfile.TemporaryDirectory() as d:
        root = Path(d)
        src = root / "contracts.ts"
        src.write_text(
            (
                "interface A { a: string }\n"
                "interface B { b: number }\n"
                "export interface Combined extends A, B {\n"
                "  id: string\n"
                "}\n\n"
                "export class Formatter {\n"
                "  format(x: string): string;\n"
                "  format(x: number): string;\n"
                "  format(x: string | number): string {\n"
                "    return String(x)\n"
                "  }\n"
                "}\n"
            ),
            encoding="utf-8",
        )
        entities = parse_typescript_module(root, root, ignore_patterns=())
        by_id = {e.code_id: e for e in entities}
        assert "contracts:Combined" in by_id
        assert "id" in by_id["contracts:Combined"].args
        assert "contracts:Formatter.format" in by_id
        assert by_id["contracts:Formatter.format"].args == ["x"]


def test_parse_doc_file_malformed_anchor() -> None:
    import tempfile
    from livedoc.parsers.doc_parser import parse_doc_file

    with tempfile.NamedTemporaryFile(mode="w", suffix=".md", delete=False) as f:
        f.write('<!-- livedoc: code_id = "" -->\n## Empty\n')
        path = Path(f.name)
    try:
        fragments = parse_doc_file(path, path.parent)
        assert len(fragments) == 0
    finally:
        path.unlink()



"""Report generation for outdated documentation fragments."""

from __future__ import annotations

import json
from pathlib import Path

from livedoc.core.graph import DocFragment
from livedoc.core.signatures import CodeEntity


def _format_change(old_sig: str | None, new_sig: str | None) -> str:
    """Format signature diff for report."""
    if old_sig is None and new_sig:
        return f"Added: {new_sig}"
    if old_sig and new_sig is None:
        return f"Removed: {old_sig}"
    if old_sig and new_sig:
        return f"{old_sig}  ->  {new_sig}"
    return "Signature changed"


def _code_location(
    code_id: str,
    entities_by_id: dict[str, CodeEntity] | None,
    project_root: Path | None,
) -> tuple[str | None, int | None]:
    """Return (relative path as str, line) for code_id, or (None, None) if unknown."""
    if not entities_by_id or code_id not in entities_by_id:
        return None, None
    ent = entities_by_id[code_id]
    path = ent.file_path
    try:
        if project_root is not None:
            path = path.resolve().relative_to(project_root.resolve())
    except ValueError:
        pass
    return str(path).replace("\\", "/"), ent.line


def _code_change_entry(
    code_id: str,
    sig_pair: tuple[str | None, str | None],
    entities_by_id: dict[str, CodeEntity] | None,
    project_root: Path | None,
) -> dict:
    old_sig, new_sig = sig_pair
    code_file, code_line = _code_location(code_id, entities_by_id, project_root)
    return {
        "code_id": code_id,
        "old_sig": old_sig,
        "new_sig": new_sig,
        "diff": _format_change(old_sig, new_sig),
        "code_file": code_file,
        "code_line": code_line,
    }


def _unknown_anchors_json(unknown_refs: list[tuple[str, DocFragment]]) -> list[dict]:
    return [
        {
            "code_id": code_id,
            "doc_fragment_id": frag.doc_fragment_id,
            "file": str(frag.file_path),
            "line": frag.line_start,
            "heading": frag.heading,
        }
        for code_id, frag in unknown_refs
    ]


def report_outdated(
    outdated: list[DocFragment],
    *,
    changes: dict[str, tuple[str | None, str | None]] | None = None,
    unknown_refs: list[tuple[str, DocFragment]] | None = None,
    entities_by_id: dict[str, CodeEntity] | None = None,
    project_root: Path | None = None,
    verbose: bool = True,
    output_format: str = "text",
) -> str:
    """
    Build report: which doc fragments are outdated and what changed.
    changes: code_id -> (old_sig, new_sig)
    unknown_refs: (code_id, fragment) for anchors pointing to missing code
    entities_by_id: current code entities (for code file:line in report)
    project_root: used to show paths relative to project
    output_format: "text" or "json"
    """
    changes = changes or {}
    unknown_refs = unknown_refs or []
    has_unknown = len(unknown_refs) > 0
    has_outdated = len(outdated) > 0

    if output_format == "json":
        payload = {
            "ok": not has_outdated and not has_unknown,
            "outdated": [
                {
                    "doc_fragment_id": f.doc_fragment_id,
                    "file": str(f.file_path),
                    "line": f.line_start,
                    "heading": f.heading,
                    "code_changes": [
                        _code_change_entry(
                            cid,
                            changes.get(cid, (None, None)),
                            entities_by_id,
                            project_root,
                        )
                        for cid in f.code_ids
                    ],
                }
                for f in outdated
            ],
            "unknown_anchors": _unknown_anchors_json(unknown_refs),
        }
        return json.dumps(payload, indent=2, ensure_ascii=False)

    lines: list[str] = []

    if has_outdated:
        lines.extend(
            [
                "Possibly outdated documentation (code changed):",
                "",
            ]
        )
        for f in outdated:
            lines.append(f"  * {f.doc_fragment_id}")
            lines.append(f"    File: {f.file_path}, line ~{f.line_start}")
            if f.heading:
                lines.append(f"    Section: {f.heading}")
            for code_id in f.code_ids:
                old_sig, new_sig = changes.get(code_id, (None, None))
                diff = _format_change(old_sig, new_sig)
                lines.append(f"    [{code_id}]  {diff}")
                code_file, code_line = _code_location(code_id, entities_by_id, project_root)
                if code_file is not None and code_line is not None:
                    lines.append(f"    Code: {code_file}:{code_line}")
                elif new_sig is None and old_sig is not None:
                    lines.append("    Code: (symbol removed from codebase)")
            if verbose:
                lines.append("    Suggestion: update parameter/return value descriptions in the docs.")
            lines.append("")

    if has_unknown:
        lines.extend(
            [
                "Unknown code_id references (not found in parsed project code):",
                "",
            ]
        )
        for code_id, f in unknown_refs:
            lines.append(f"  * [{code_id}]")
            lines.append(f"    Doc: {f.doc_fragment_id}")
            lines.append(f"    File: {f.file_path}, line ~{f.line_start}")
            if f.heading:
                lines.append(f"    Section: {f.heading}")
            if verbose:
                lines.append(
                    "    Fix: correct the anchor or add the symbol to the codebase; "
                    "check ignore patterns if the file is excluded."
                )
            lines.append("")

    if not has_outdated and not has_unknown:
        return "Documentation is up to date: no changes detected in linked code."

    return "\n".join(lines).rstrip() + "\n"



"""End-to-end CLI tests built on small fixture projects."""

from __future__ import annotations

import json
import shutil
import sys
from pathlib import Path

import pytest

from livedoc.core.signatures import CodeSignatures


FIXTURES_ROOT = Path(__file__).resolve().parent / "fixtures"


def _copy_fixture(tmp_path: Path, name: str) -> Path:
    project_root = tmp_path / name
    shutil.copytree(FIXTURES_ROOT / name, project_root)
    return project_root


def _run_main(monkeypatch: pytest.MonkeyPatch, *args: str | Path) -> int:
    from livedoc.cli import main

    monkeypatch.setattr(sys, "argv", ["livedoc", *[str(arg) for arg in args]])
    return main()


def test_main_e2e_first_run_then_second_run_ok(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    project_root = _copy_fixture(tmp_path, "e2e_project")

    rc_first = _run_main(monkeypatch, project_root, "--quiet")
    out_first = capsys.readouterr()
    assert rc_first == 0
    assert out_first.out == ""

    sig_path = project_root / ".livedoc" / "code_signatures.json"
    stored = CodeSignatures.load(sig_path)
    assert stored is not None
    assert "math:add" in stored.signatures
    assert "web.service:render" in stored.signatures
    assert "calc:Multiply" in stored.signatures

    rc_second = _run_main(monkeypatch, project_root, "--quiet")
    out_second = capsys.readouterr()
    assert rc_second == 0
    assert out_second.out == ""


def test_main_e2e_first_run_json_is_machine_readable(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    project_root = _copy_fixture(tmp_path, "e2e_project")

    rc = _run_main(monkeypatch, project_root, "--format", "json")
    out = capsys.readouterr()
    assert rc == 0
    payload = json.loads(out.out)
    assert payload["ok"] is True
    assert payload["outdated"] == []
    assert payload["unknown_anchors"] == []


def test_main_e2e_detects_change_then_allows_update(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    project_root = _copy_fixture(tmp_path, "e2e_project")
    assert _run_main(monkeypatch, project_root, "--quiet") == 0
    _ = capsys.readouterr()

    ts_file = project_root / "web" / "service.ts"
    ts_file.write_text(
        (
            "export function render(name: string, locale: string): string {\n"
            "  return `${locale}:${name}`;\n"
            "}\n"
        ),
        encoding="utf-8",
    )

    rc_changed = _run_main(monkeypatch, project_root, "--format", "json", "--quiet")
    out_changed = capsys.readouterr()
    assert rc_changed == 1
    payload = json.loads(out_changed.out)
    assert payload["ok"] is False
    assert payload["unknown_anchors"] == []
    assert len(payload["outdated"]) == 1
    code_change = payload["outdated"][0]["code_changes"][0]
    assert code_change["code_id"] == "web.service:render"
    assert code_change["code_file"] == "web/service.ts"
    assert code_change["new_sig"] == "render(name, locale) -> string"

    rc_update = _run_main(monkeypatch, project_root, "--update", "--quiet")
    out_update = capsys.readouterr()
    assert rc_update == 1
    assert "web.service:render" in out_update.out

    rc_after_update = _run_main(monkeypatch, project_root, "--quiet")
    out_after_update = capsys.readouterr()
    assert rc_after_update == 0
    assert out_after_update.out == ""


def test_main_e2e_reports_unknown_anchor_and_still_saves_signatures(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    project_root = _copy_fixture(tmp_path, "e2e_project")
    docs_file = project_root / "docs" / "api.md"
    docs_file.write_text(
        docs_file.read_text(encoding="utf-8")
        + '\n<!-- livedoc: code_id = "missing:Symbol" -->\n## Missing\n\nBroken reference.\n',
        encoding="utf-8",
    )

    rc = _run_main(monkeypatch, project_root, "--format", "json")
    out = capsys.readouterr()
    assert rc == 1
    payload = json.loads(out.out)
    assert payload["ok"] is False
    assert payload["outdated"] == []
    assert payload["unknown_anchors"][0]["code_id"] == "missing:Symbol"

    sig_path = project_root / ".livedoc" / "code_signatures.json"
    assert sig_path.exists()
